# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Optimize performance of gRPC communications.

GrpcCallBatcher may be used to reduce the number of remote communications
by batching together API commands that do not require return values.
"""

from enum import Enum
import os
import typing

from ..internal.apiutil import SupportsDeleteCallback
from .exceptions import GrpcUtilitiesError

try:
    from ..internal.AgGrpcServices_pb2 import BatchedInvokeRequest, InvokeRequest
    from ..internal.grpcutil import GrpcClient, GrpcInterfaceFuture, GrpcInterfacePimpl
    _DEFAULT_BATCH_DISABLE = False
except ImportError:
    _DEFAULT_BATCH_DISABLE = True

class GrpcAuthenticationMode(Enum):
    """Specify the method of client-server authentication to use for gRPC."""

    SINGLE_USER = 0
    """Ensure client and server user match. Only available on Windows."""
    UNIX_DOMAIN_SOCKET = 1
    """Ensure client has permission to access server socket file using Unix domain socket names. Only available on Linux."""
    MUTUAL_TLS = 2
    """Perform mutual TLS by providing paths for client and server certificate, key, and certificate authority files."""
    INSECURE = 3
    """Disable authentication, not recommended. Before using this method, consider a secure connection."""
    DEFAULT = UNIX_DOMAIN_SOCKET if os.name != "nt" else SINGLE_USER
    """The default mode for starting or attaching to an STK application. SINGLE_USER on Windows, UNIX_DOMAIN_SOCKET on Linux."""

class GrpcCallBatcher(object):
    """
    A class used to batch together API calls to optimize performance.

    Activating batching will cause the normal API exception behavior to be
    altered. Exceptions from one command may appear asynchronously. Therefore
    it is not recommended to use call batching while building and debugging,
    but rather as a performance optimization.

    Only calls that do not return a value may be batched together,
    such as set-property requests and methods without a return value.
    Any method that has a return value (including get-property requests)
    will automatically execute any previously batched commands before the
    method with a return value is executed.

    Therefore, to reduce the number of remote API requests and improve
    performance, code must be organized to group together commands that
    do not have a return value. Call chaining will interrupt a batch request
    because of the get-property command within the chain. E.g.:

        root.CurrentScenario.ShortDescription = short_description
        root.CurrentScenario.LongDescription = long_description

    will not be batched together because the call to `CurrentScenario` will
    get the scenario via an API call. These commands may be batched by
    factoring out the call chaining:

        scen = root.CurrentScenario
        scen.ShortDescription = short_description
        scen.LongDescription = long_description

    This class may be used via the explicit commands or by using the "with"
    statement to batch together the commands within the statement block.
    e.g.

        call_batcher = stk.NewGrpcCallBatcher()
        with call_batcher:
            facility1.LocalTimeOffset = 1.0
            facility1.HeightAboveGround = 10.0
            facility1.UseLocalTimeOffset = True
            facility1.ResetAzElMask()
    """

    _default_max_batch_size = 10000
    _disable_batching = _DEFAULT_BATCH_DISABLE

    def __init__(self, disable_batching=False):
        """Construct an object of type GrpcCallBatcher."""
        self._initialized = False
        self._max_batch = GrpcCallBatcher._default_max_batch_size
        self._disable_batching = GrpcCallBatcher._disable_batching or disable_batching
        self._reset()

    def _reset(self):
        self._next_future_id = 1
        self._unbound_futures = {}

    def _private_init(self, client:"GrpcClient", max_batch:int=None) -> None:
        self._initialized = True
        self._client = client
        self._requests = BatchedInvokeRequest()
        self._batching = False
        if max_batch is not None:
            if max_batch > GrpcCallBatcher._default_max_batch_size:
                raise GrpcUtilitiesError(f"Batch size cannot exceed {GrpcCallBatcher._default_max_batch_size} due to gRPC message size restrictions.")
            self._max_batch = max_batch

    def __enter__(self):
        """Use GrpcCallBatcher with the "with" statement to activate batching."""
        if not self._disable_batching:
            self.start_batching()
        return self

    def __exit__(self, type, value, tb):
        """Use GrpcCallBatcher with the "with" statement to deactivate batching."""
        if not self._disable_batching:
            self.stop_batching()
        return False

    def start_batching(self) -> None:
        """Explicitly start batching until stop_batching() is called."""
        if not self._disable_batching and not self._batching:
            if not self._initialized:
                raise RuntimeError("The GrpcCallBatcher should be obtained from the STK application rather than constructed directly.")
            GrpcClient.register_call_batcher(self)
            self._batching = True

    def execute_batch(self) -> None:
        """Explicitly execute any queued batch commands."""
        if not self._disable_batching and self._batching:
            bindings = GrpcClient.execute_call_batcher(self)
            if bindings is not None:
                for binding in bindings:
                    GrpcClient.bind_future(self, self._unbound_futures[binding.obj_future_id], binding)
            self._reset()

    def stop_batching(self) -> None:
        """Explicitly stop batching."""
        if not self._disable_batching and self._initialized:
            self.execute_batch()
            GrpcClient.unregister_call_batcher(self)
            self._batching = False

    def _enqueue_batch_request(self, request:"InvokeRequest") -> None:
        self._requests.requests.append(request)

    def _ready_for_invoke(self) -> bool:
        if len(self._requests.requests) >= self._max_batch:
            return True
        return False

    def _get_requests_for_invoke(self) -> "BatchedInvokeRequest":
        batch = self._requests
        self._requests = BatchedInvokeRequest()
        active_futures = self._get_active_futures()
        if len(active_futures) > 0:
            batch.future_ids_to_return.extend(active_futures)
        return batch

    def _get_active_futures(self) -> typing.List[int]:
        active_list = []
        for future_id in self._unbound_futures:
            if self._unbound_futures[future_id].active:
                active_list.append(future_id)
        return active_list

    @staticmethod
    def _bypass_future_creation(source_obj:typing.Any, future_provider:typing.Union[typing.Callable, property], *args) -> typing.Any:
        if callable(future_provider):
            if hasattr(source_obj, future_provider.__name__):
                return future_provider(source_obj, *args)
        elif type(future_provider) is property:
            attr_name = None
            for superclass in reversed(source_obj.__class__.mro()):
                if hasattr(superclass, "_property_names"):
                    if future_provider in superclass._property_names:
                        attr_name = superclass._property_names[future_provider]
                        break
            if attr_name is None:
                raise GrpcUtilitiesError("Cannot create gRPC future; incorrect type.")
            return getattr(source_obj, attr_name)

    def create_future(self, source_obj:typing.Any, future_provider:typing.Union[typing.Callable, property], future_type:typing.Any, *args) -> typing.Any:
        """
        Create an object of type future_type that supports batching operations.

        source_obj is an STK Object Model type, e.g. STKObjectRoot.
        future_provider is a member method or property of source_obj, e.g. STKObjectRoot.CurrentScenario.
        future_type is the STK Object Model type that is returned from future_provider, e.g. Scenario.
        args are the arguments passed to future_provider if applicable.
        """
        if self._disable_batching:
            return GrpcCallBatcher._bypass_future_creation(source_obj, future_provider, *args)
        if not self._batching:
            raise GrpcUtilitiesError("Batcher must be active to create futures.")
        if not callable(future_type):
            raise GrpcUtilitiesError("Future class type must be a full STK Object type (e.g. Scenario, not Scenario).")
        future = future_type()
        if not isinstance(future, SupportsDeleteCallback):
            raise GrpcUtilitiesError("Future class type must be a full STK Object type (e.g. Scenario, not Scenario).")
        intf_proxy = GrpcInterfaceFuture(self, self._next_future_id, source_obj, future_provider, *args)
        intf_pimpl = GrpcInterfacePimpl(intf_proxy)
        future._private_init(intf_pimpl)
        future._add_delete_callback(intf_pimpl.deactivate)
        self._unbound_futures[self._next_future_id] = intf_pimpl
        self._next_future_id += 1
        return future

def _get_authentication_mode_string(authentication_mode:GrpcAuthenticationMode):
    if authentication_mode == GrpcAuthenticationMode.UNIX_DOMAIN_SOCKET:
        return "uds"
    elif authentication_mode == GrpcAuthenticationMode.MUTUAL_TLS:
        return "mtls"
    elif authentication_mode == GrpcAuthenticationMode.INSECURE:
        return "insecure"
    else:
        return "uds" if os.name != "nt" else "single-user"

def _validate_authentication_mode(authentication_mode:GrpcAuthenticationMode, grpc_host:str):
    if authentication_mode == GrpcAuthenticationMode.SINGLE_USER:
        if os.name != "nt":
            raise RuntimeError("Single user authentication is not supported on Linux. Choose another authentication method.")
        if grpc_host not in ["localhost", "127.0.0.1"]:
            raise RuntimeError("Single user authentication is not supported with remote server. Choose another authentication method.")
    elif authentication_mode == GrpcAuthenticationMode.UNIX_DOMAIN_SOCKET:
        if os.name == "nt":
            raise RuntimeError("UDS is not supported on Windows. Choose another authentication method.")


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Recording of API calls."""

import json
import logging
from pathlib import Path
import sys
from typing import Any

from .call_record import CallRecord


class Recording:
    """Group of calls recorded during one or multiple sessions of the recorder."""

    def __init__(self, root_directory_path):
        self.call_records = list()
        self.root_directory_path = root_directory_path

    def add(
        self,
        filename,
        type_name,
        member_name,
        lineno,
        end_lineno,
        col_offset,
        end_col_offset,
    ):
        """Add a new call record to this recording."""
        try:
            relative_filename = str(Path(filename).relative_to(self.root_directory_path))
        except ValueError:
            logging.warning(f"{filename} is not under {self.root_directory_path}")
        else:
            new_record = CallRecord(
                relative_filename,
                type_name,
                member_name,
                lineno,
                end_lineno,
                col_offset,
                end_col_offset,
            )
            if new_record not in self.call_records:
                self.call_records.append(new_record)

    def sort_call_records(self):
        """Sort the call recorded previously added."""
        self.call_records = sorted(list(set(self.call_records)))

    def save(self, file_name, description=None):
        """Save this recording to the specified file."""
        sorted_call_records = sorted(self.call_records)

        class CallRecordEncoder(json.JSONEncoder):
            def default(self, o: Any):
                if isinstance(o, CallRecord):
                    return {
                        "filename": o.filename,
                        "lineno": o.lineno,
                        "end_lineno": o.end_lineno,
                        "col_offset": o.col_offset,
                        "end_col_offset": o.end_col_offset,
                        "type_name": o.type_name,
                        "member_name": o.member_name,
                        }
                return super(CallRecordEncoder, self).default(o)

        recording = {
            "root_directory": str(self.root_directory_path.resolve()),
            "calls": sorted_call_records,
        }
        if description is not None:
            recording["command"] = description

        with Path(file_name).open(mode="w") as f:
            json.dump(recording, f, cls=CallRecordEncoder, indent=4)

    @staticmethod
    def load_from_recordings_in_directory(recordings_directory):
        """Load a new recording from the files in the specified directory."""
        recording_files = Path(recordings_directory).glob("*.json")

        recording = Recording("")
        root_directory = None

        for recording_file in recording_files:
            logging.debug(f"Processing {recording_file}")

            print(recording_file)

            with Path(recording_file).open(mode="r") as f:
                single_recording = json.load(f)

                current_file_root_directory = single_recording["root_directory"]
                if root_directory is None:
                    root_directory = current_file_root_directory
                elif root_directory != current_file_root_directory:
                    logging.error("Inconsistent recording files based on different root directories!")
                    sys.exit(-3)

                for call in single_recording["calls"]:
                    recording.add(
                        call.get("filename"),
                        call.get("type_name"),
                        call.get("member_name"),
                        int(call.get("lineno")),
                        int(call.get("end_lineno")),
                        int(call.get("col_offset")),
                        int(call.get("end_col_offset")),
                    )

        if root_directory is not None:
            recording.root_directory_path = Path(root_directory).resolve()
        recording.sort_call_records()
        return recording

    def get_files_to_edit(self):
        """Get the files that must be edited based on this recording."""
        files_to_edit = set(
            [str((self.root_directory_path / record.filename).resolve()) for record in self.call_records]
        )
        return files_to_edit

    def get_change_for_file_region(self, filename, member_name, lineno, end_lineno, col_offset, end_col_offset):
        """Find the change corresponding to the specified region."""
        relative_filename = str(Path(filename).relative_to(self.root_directory_path))
        changes = [
            record
            for record in self.call_records
            if record.filename == relative_filename
            and record.member_name == member_name
            and record.lineno <= lineno
            and record.end_lineno >= end_lineno
            and record.col_offset <= col_offset
            and record.end_col_offset >= end_col_offset
        ]

        if len(changes) >= 1:
            return changes[0]
        else:
            return None



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import annotations

import os
import typing
import copy

from ctypes import c_void_p, c_longlong, c_ulonglong, c_int, c_uint, c_ulong, c_ushort, c_short, c_ubyte, c_wchar_p, c_double, c_float, c_bool
from ctypes import POINTER, Structure, Union, byref, cast, pointer

from .apiutil import OutArg, GcDisabler

###############################################################################
#   COM Types
###############################################################################
BSTR = c_wchar_p
BYTE = c_ubyte
CHAR = c_ubyte
DOUBLE = c_double
DWORD = c_ulong
INT = c_int
LONG = c_int #don't use c_long because of cross-platform size differences
FLOAT = c_float
HRESULT = c_int
LONGLONG = c_longlong
LPCOLESTR = c_wchar_p
LPCVOID = c_void_p
LPOLESTR = c_wchar_p
LPVOID = c_void_p
LPCWSTR = c_wchar_p
LPWSTR = c_wchar_p
OLESTR = c_wchar_p
PVOID = c_void_p
SHORT = c_short
UINT = c_uint
ULONG = c_uint #don't use c_long because of cross-platform size differences
ULONGLONG = c_ulonglong
USHORT = c_ushort
VARIANT_BOOL = c_short
VARTYPE = c_int
WORD = c_ushort
BOOL = c_bool

DATE = DOUBLE
DISPID = LONG
LCID = DWORD
OLE_COLOR = ULONG
OLE_HANDLE = UINT
OLE_XPOS_PIXELS = LONG
OLE_YPOS_PIXELS = LONG
LPSAFEARRAY = PVOID
LPSTREAM = PVOID


###############################################################################
#   Constants
###############################################################################
# CLSCTX Enumeration
CLSCTX_INPROC_SERVER = 1
CLSCTX_LOCAL_SERVER = 0x4

# COINIT enumeration
COINIT_APARTMENTTHREADED = 0x2

# HRESULT values
S_OK                = 0
E_NOTIMPL           = 0x80004001
E_NOINTERFACE       = 0x80004002
CO_E_NOTINITIALIZED = 0x800401F0

# Numeric Limits
ULLONG_MAX =  0xffffffffffffffff
LLONG_MAX  =  9223372036854775807
LLONG_MIN  = -9223372036854775808
LONG_MAX   =  2147483647
LONG_MIN   = -2147483648

# VARIANT_BOOL
VARIANT_FALSE = VARIANT_BOOL(0)
VARIANT_TRUE = VARIANT_BOOL(-1)

# Variant Types
VT_EMPTY           = 0x0000
VT_NULL            = 0x0001
VT_I2              = 0x0002
VT_I4              = 0x0003
VT_R4              = 0x0004
VT_R8              = 0x0005
VT_CY              = 0x0006
VT_DATE            = 0x0007
VT_BSTR            = 0x0008
VT_DISPATCH        = 0x0009
VT_ERROR           = 0x000A
VT_BOOL            = 0x000B
VT_VARIANT         = 0x000C
VT_UNKNOWN         = 0x000D
VT_DECIMAL         = 0x000E
VT_I1              = 0x0010
VT_UI1             = 0x0011
VT_UI2             = 0x0012
VT_UI4             = 0x0013
VT_I8              = 0x0014
VT_UI8             = 0x0015
VT_INT             = 0x0016
VT_UINT            = 0x0017
VT_VOID            = 0x0018
VT_HRESULT         = 0x0019
VT_PTR             = 0x001A
VT_SAFEARRAY       = 0x001B
VT_CARRAY          = 0x001C
VT_USERDEFINED     = 0x001D
VT_LPSTR           = 0x001E
VT_LPWSTR          = 0x001F
VT_INT_PTR         = 0x0025
VT_UINT_PTR        = 0x0026
VT_ARRAY           = 0x2000
VT_BYREF           = 0x4000


###############################################################################
#   COM Structures
###############################################################################
class GUID(Structure):
    _fields_ = [("Data1", BYTE*4), ("Data2", BYTE*2), ("Data3", BYTE*2), ("Data4", BYTE*8)]
    def __init__(self, name=None):
        if name is not None:
            if OLE32Lib.CLSIDFromString is None:
                OLE32Lib._initialize()
            OLE32Lib.CLSIDFromString(str(name), byref(self))

    @staticmethod
    def from_registry_format(data: str) -> "GUID":
        if len(data) != 38 or data[0] != "{" or data[-1] != "}":
            raise ValueError(f"{data} not a GUID in registry format")

        guid = GUID()

        data_bytes = bytes.fromhex(data[1:-1].replace('-', ''))

        guid.Data1 = (BYTE * 4).from_buffer_copy(data_bytes[3::-1])

        guid.Data2[0] = data_bytes[5]
        guid.Data2[1] = data_bytes[4]

        guid.Data3[0] = data_bytes[7]
        guid.Data3[1] = data_bytes[6]

        guid.Data4 = (BYTE * 8).from_buffer_copy(data_bytes[8:16])

        return guid

    @staticmethod
    def from_data_pair(data:tuple) -> "GUID":
        guid_union = _guid_union(data)
        guid = copy.deepcopy(guid_union.guid)
        return guid

    def as_data_pair(self) -> tuple:
        guid_union = _guid_union(self)
        return (guid_union.data[0], guid_union.data[1])

    def __eq__(self, other):
        are_equal = True
        if self.Data1[0] != other.Data1[0]: are_equal = False
        if self.Data1[1] != other.Data1[1]: are_equal = False
        if self.Data1[2] != other.Data1[2]: are_equal = False
        if self.Data1[3] != other.Data1[3]: are_equal = False
        if self.Data2[0] != other.Data2[0]: are_equal = False
        if self.Data2[1] != other.Data2[1]: are_equal = False
        if self.Data3[0] != other.Data3[0]: are_equal = False
        if self.Data3[1] != other.Data3[1]: are_equal = False
        if self.Data4[0] != other.Data4[0]: are_equal = False
        if self.Data4[1] != other.Data4[1]: are_equal = False
        if self.Data4[2] != other.Data4[2]: are_equal = False
        if self.Data4[3] != other.Data4[3]: are_equal = False
        if self.Data4[4] != other.Data4[4]: are_equal = False
        if self.Data4[5] != other.Data4[5]: are_equal = False
        if self.Data4[6] != other.Data4[6]: are_equal = False
        if self.Data4[7] != other.Data4[7]: are_equal = False
        return are_equal

    def __str__(self):
        return "{{{:02X}{:02X}{:02X}{:02X}-{:02X}{:02X}-{:02X}{:02X}-{:02X}{:02X}-{:02X}{:02X}{:02X}{:02X}{:02X}{:02X}}}".format(
            self.Data1[3],
            self.Data1[2],
            self.Data1[1],
            self.Data1[0],
            self.Data2[1],
            self.Data2[0],
            self.Data3[1],
            self.Data3[0],
            self.Data4[0],
            self.Data4[1],
            self.Data4[2],
            self.Data4[3],
            self.Data4[4],
            self.Data4[5],
            self.Data4[6],
            self.Data4[7])

class _guid_union(Union):
    _fields_ = [("guid", GUID), ("data", ULONGLONG*2)]
    def __init__(self, iid:typing.Union[GUID, tuple]):
        if type(iid) == tuple:
            self.data[0] = iid[0]
            self.data[1] = iid[1]
        else:
            self.guid = iid

IID = GUID
REFIID = POINTER(IID)

class varUnion(Union):
    # GCC throws an error when trying to marshall ctypes.Union to C++ methods. Therefore the Variant class
    # only has a raw buffer that is equivalent in size to the varUnion.
    _fields_ = [
                ("buffer", BYTE*16),
                ("llVal", LONGLONG), #VT_I8
                ("lVal", LONG), #VT_I4
                ("bVal", BYTE), #VT_UI1
                ("iVal", SHORT), #VT_I2
                ("fltVal", FLOAT), #VT_R4
                ("dblVal", DOUBLE), #VT_R8
                ("boolVal", VARIANT_BOOL), #VT_BOOL
                ("bstrVal", BSTR), #VT_BSTR
                ("punkVal", PVOID), #VT_UNKNOWN
                ("parray", LPSAFEARRAY), #VT_ARRAY
                ("pbVal", POINTER(BYTE)), #VT_UI1|VT_BYREF
                ("piVal", POINTER(SHORT)), #VT_I2|VT_BYREF
                ("plVal", POINTER(LONG)), #VT_I4|VT_BYREF
                ("pllVal", POINTER(LONGLONG)), #VT_I8|VT_BYREF
                ("pfltVal", POINTER(FLOAT)), #VT_R4|VT_BYREF
                ("pdblVal", POINTER(DOUBLE)), #VT_R8|VT_BYREF
                ("pboolVal", POINTER(VARIANT_BOOL)), #VT_BOOL|VT_BYREF
                ("pbstrVal", POINTER(BSTR)), #VT_BSTR|VT_BYREF
                ("ppunkVal", POINTER(PVOID)), #VT_UNKNOWN|VT_BYREF
                ("pparray", POINTER(LPSAFEARRAY)), #VT_ARRAY|VT_BYREF
                ("cVal", CHAR), #VT_I1
                ("uiVal", USHORT), #VT_UI2
                ("ulVal", ULONG), #VT_UI4
                ("ullVal", ULONGLONG), #VT_UI8
                ("intVal", INT), #VT_INT
                ("uintVal", UINT), #VT_UINT
                ("pcVal", POINTER(CHAR)), #VT_I1|VT_BYREF
                ("puiVal", POINTER(USHORT)), #VT_UI2|VT_BYREF
                ("pulVal", POINTER(ULONG)), #VT_UI4|VT_BYREF
                ("pullVal", POINTER(ULONGLONG)), #VT_UI8|VT_BYREF
                ("pintVal", POINTER(INT)), #VT_INT|VT_BYREF
                ("puintVal", POINTER(UINT)), #VT_UINT|VT_BYREF
                ("pvarVal", POINTER(PVOID)), #VT_VARIANT|VT_BYREF
                ("scode", HRESULT), #VT_ERROR
                ("pscode", POINTER(HRESULT)), #VT_ERROR|VT_BYREF
                ("pdispVal", PVOID), #VT_DISPATCH
                ("ppdispVal", POINTER(PVOID)) #VT_DISPATCH|VT_BYREF
               ]

class Variant(Structure):
    # Copy a varUnion into the buffer to get the correct data.
    _fields_ = [("vt", WORD), ("wReserved1", WORD), ("wReserved2", WORD), ("wReserved3", WORD), ("buffer", BYTE*16)]
    def __init__(self):
        # To initialize Variant from python data, use marshall.VARIANT_from_python_data
        pass

class SafearrayBound(Structure):
    _fields_ = [("cElements", ULONG), ("lLbound", LONG)]

class DispParams(Structure):
    _fields_ = [("rgvarg", POINTER(Variant)), ("rgdispidNamedArgs", POINTER(DISPID)), ("cArgs", UINT), ("cNamedArgs", UINT)]

class ExcepInfo(Structure):
    _fields_ = [
                ("wCode", WORD),
                ("wReserved", WORD),
                ("bstrSource", BSTR),
                ("bstrDescription", BSTR),
                ("bstrHelpFile", BSTR),
                ("dwHelpContext", DWORD),
                ("pvReserved", POINTER(ULONG)),
                ("pfnDeferredFillIn", POINTER(ULONG)),
                ("scode", HRESULT)
            ]

###############################################################################
#   COM Methods
###############################################################################

if os.name == "nt":
    from ctypes import WINFUNCTYPE
else:
    from ctypes import CFUNCTYPE
    WINFUNCTYPE = CFUNCTYPE

class OLE32Lib:

    _handle = None

    CLSIDFromString     = None
    CLSIDFromProgID     = None
    CoCreateInstance    = None
    CoInitializeEx      = None
    CoTaskMemFree       = None
    CoUninitialize      = None
    StringFromCLSID     = None

    if os.name=="nt":

        CoMarshalInterThreadInterfaceInStream = None
        CoGetInterfaceAndReleaseStream        = None
        CoReleaseMarshalData                  = None

        CreateClassMoniker    = None
        GetRunningObjectTable = None
        CreateBindCtx         = None
        CoGetMalloc           = None

    def _initialize():

        if OLE32Lib._handle is not None:
            return

        if os.name == "nt":
            from ctypes import windll
            OLE32Lib._handle = windll.ole32
        else:
            from ctypes import cdll
            OLE32Lib._handle = cdll.LoadLibrary("libagxcom.so")

        OLE32Lib.CLSIDFromString     = WINFUNCTYPE(HRESULT, LPCWSTR, POINTER(GUID))(("CLSIDFromString", OLE32Lib._handle), ((1, "lpsz"), (1, "pclsid")))
        OLE32Lib.CLSIDFromProgID     = WINFUNCTYPE(HRESULT, LPCWSTR, POINTER(GUID))(("CLSIDFromProgID", OLE32Lib._handle), ((1, "lpszProgID"), (1, "lpclsid")))
        OLE32Lib.CoCreateInstance    = WINFUNCTYPE(HRESULT, POINTER(GUID), LPVOID, DWORD, POINTER(GUID), POINTER(LPVOID))(("CoCreateInstance", OLE32Lib._handle),
                                       ((1, "rclsid"), (1, "pUnkOuter"), (1, "dwClsContext"), (1, "riid"), (1, "ppv")))
        OLE32Lib.CoInitializeEx      = WINFUNCTYPE(HRESULT, c_void_p, DWORD)(("CoInitializeEx", OLE32Lib._handle), ((1, "pvReserved"), (1, "dwCoInit")))
        OLE32Lib.CoTaskMemFree       = WINFUNCTYPE(None, LPVOID)(("CoTaskMemFree", OLE32Lib._handle), ((1, "pv"),))
        OLE32Lib.CoUninitialize      = WINFUNCTYPE(None)(("CoUninitialize", OLE32Lib._handle))
        OLE32Lib.StringFromCLSID     = WINFUNCTYPE(HRESULT, POINTER(GUID), POINTER(LPOLESTR))(("StringFromCLSID", OLE32Lib._handle), ((1, "rclsid"), (1, "lplpsz")))

        if os.name=="nt":

            OLE32Lib.CoMarshalInterThreadInterfaceInStream = WINFUNCTYPE(HRESULT, REFIID, PVOID, POINTER(LPSTREAM))(("CoMarshalInterThreadInterfaceInStream", OLE32Lib._handle), ((1, "riid"), (1, "pUnk"), (1, "ppStm")))
            OLE32Lib.CoGetInterfaceAndReleaseStream        = WINFUNCTYPE(HRESULT, LPSTREAM, REFIID, POINTER(PVOID))(("CoGetInterfaceAndReleaseStream", OLE32Lib._handle), ((1, "pStm"), (1, "iid"), (1, "ppv")))
            OLE32Lib.CoReleaseMarshalData                  = WINFUNCTYPE(HRESULT, LPSTREAM)(("CoReleaseMarshalData", OLE32Lib._handle), ((1, "pStm"),))

            OLE32Lib.CreateClassMoniker    = WINFUNCTYPE(HRESULT, GUID, POINTER(LPVOID))(("CreateClassMoniker", OLE32Lib._handle), ((1, "rclsid"), (1, "ppmk")))
            OLE32Lib.GetRunningObjectTable = WINFUNCTYPE(HRESULT, DWORD, POINTER(LPVOID))(("GetRunningObjectTable", OLE32Lib._handle), ((1, "dwReserved"), (1, "pprot")))
            OLE32Lib.CreateBindCtx         = WINFUNCTYPE(HRESULT, DWORD, POINTER(LPVOID))(("CreateBindCtx", OLE32Lib._handle), ((1, "dwReserved"), (1, "ppbc")))
            OLE32Lib.CoGetMalloc           = WINFUNCTYPE(HRESULT, DWORD, POINTER(LPVOID))(("CoGetMalloc", OLE32Lib._handle), ((1, "dwMemContext"), (1, "ppMalloc")))

class OLEAut32Lib:

    _handle = None

    GetActiveObject     = None
    GetErrorInfo        = None
    SafeArrayCreate     = None
    SafeArrayDestroy    = None
    SafeArrayGetDim     = None
    SafeArrayGetLBound  = None
    SafeArrayGetUBound  = None
    SafeArrayGetVartype = None
    SafeArrayGetElement = None
    SafeArrayPutElement = None
    SysAllocString      = None
    SysFreeString       = None
    VariantClear        = None
    VariantCopy         = None
    VariantInit         = None

    def _initialize():

        if OLEAut32Lib._handle is not None:
            return

        if os.name == "nt":
            from ctypes import windll
            OLEAut32Lib._handle = windll.OleAut32
        else:
            from ctypes import cdll
            OLEAut32Lib._handle = cdll.LoadLibrary("libagxcom.so")

        OLEAut32Lib.GetActiveObject     = WINFUNCTYPE(HRESULT, POINTER(GUID), LPVOID, POINTER(LPVOID))(("GetActiveObject", OLEAut32Lib._handle),
                                       ((1, "rclsid"), (1, "pvReserved"), (1, "ppunk"))) if os.name=="nt" else None
        OLEAut32Lib.GetErrorInfo        = WINFUNCTYPE(HRESULT, DWORD, POINTER(LPVOID))(("GetErrorInfo", OLEAut32Lib._handle), ((1, "dwReserved"), (1, "ppErrorInfo")))
        OLEAut32Lib.SafeArrayAccessData = WINFUNCTYPE(HRESULT, LPSAFEARRAY, POINTER(LPVOID))(("SafeArrayAccessData", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "ppvData")))
        OLEAut32Lib.SafeArrayCreate     = WINFUNCTYPE(LPSAFEARRAY, VARTYPE, UINT, POINTER(SafearrayBound))(("SafeArrayCreate", OLEAut32Lib._handle), ((1, "vt"), (1, "cDims"), (1, "rgsabound")))
        OLEAut32Lib.SafeArrayDestroy    = WINFUNCTYPE(HRESULT, LPSAFEARRAY)(("SafeArrayDestroy", OLEAut32Lib._handle), ((1, "pSafeArray"),))
        OLEAut32Lib.SafeArrayGetDim     = WINFUNCTYPE(UINT,    LPSAFEARRAY)(("SafeArrayGetDim", OLEAut32Lib._handle), ((1, "pSafeArray"),))
        OLEAut32Lib.SafeArrayGetLBound  = WINFUNCTYPE(HRESULT, LPSAFEARRAY, UINT, POINTER(LONG))(("SafeArrayGetLBound", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "nDim"), (1, "pLBound")))
        OLEAut32Lib.SafeArrayGetUBound  = WINFUNCTYPE(HRESULT, LPSAFEARRAY, UINT, POINTER(LONG))(("SafeArrayGetUBound", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "nDim"), (1, "pUBound")))
        OLEAut32Lib.SafeArrayGetVartype = WINFUNCTYPE(HRESULT, LPSAFEARRAY, POINTER(VARTYPE))(("SafeArrayGetVartype", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "vt")))
        OLEAut32Lib.SafeArrayGetElement = WINFUNCTYPE(HRESULT, LPSAFEARRAY, POINTER(LONG), PVOID)(("SafeArrayGetElement", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "rgIndices"), (1, "pElement")))
        OLEAut32Lib.SafeArrayPutElement = WINFUNCTYPE(HRESULT, LPSAFEARRAY, POINTER(LONG), PVOID)(("SafeArrayPutElement", OLEAut32Lib._handle), ((1, "pSafeArray"), (1, "rgIndices"), (1, "pElement")))
        OLEAut32Lib.SafeArrayUnaccessData = WINFUNCTYPE(HRESULT, LPSAFEARRAY)(("SafeArrayUnaccessData", OLEAut32Lib._handle), ((1, "pSafeArray"),))
        OLEAut32Lib.SysAllocString      = WINFUNCTYPE(LPVOID, LPOLESTR)(("SysAllocString", OLEAut32Lib._handle), ((1, "psz"),))
        OLEAut32Lib.SysFreeString       = WINFUNCTYPE(None, LPOLESTR)(("SysFreeString", OLEAut32Lib._handle), ((1, "bstrString"),))
        OLEAut32Lib.VariantClear        = WINFUNCTYPE(HRESULT, POINTER(Variant))(("VariantClear", OLEAut32Lib._handle), ((1, "pVariant"),))
        OLEAut32Lib.VariantCopy         = WINFUNCTYPE(HRESULT, POINTER(Variant), POINTER(Variant))(("VariantCopy", OLEAut32Lib._handle), ((1, "pvargDest"),(1, "pvargSrc")))
        OLEAut32Lib.VariantInit         = WINFUNCTYPE(None, POINTER(Variant))(("VariantInit", OLEAut32Lib._handle), ((1, "pVariant"),))

###############################################################################
#   Helper functions
###############################################################################
def Succeeded(hr):
    return hr >= S_OK

class _CreateAgObjectLifetimeManager(object):
    """Singleton class for managing reference counts on COM interfaces."""
    _AddRef = WINFUNCTYPE(ULONG, LPVOID)
    _Release = WINFUNCTYPE(ULONG, LPVOID)
    if os.name == "nt":
        _AddRefIndex = 1
        _ReleaseIndex = 2
    else:
        _AddRefIndex = 0
        _ReleaseIndex = 1

    def __init__(self):
        self._ref_counts = dict()
        self._applications = list()

    @staticmethod
    def _release_impl(pUnk:"IUnknown"):
        """Call Release in STK."""
        _CreateAgObjectLifetimeManager._Release(pUnk._get_vtbl_entry(_CreateAgObjectLifetimeManager._ReleaseIndex))(pUnk.p)

    @staticmethod
    def _add_ref_impl(pUnk:"IUnknown"):
        """Call AddRef in STK."""
        _CreateAgObjectLifetimeManager._AddRef(pUnk._get_vtbl_entry(_CreateAgObjectLifetimeManager._AddRefIndex))(pUnk.p)

    def create_ownership(self, pUnk:"IUnknown"):
        """
        Add pUnk to the reference manager and call AddRef in STK.

        Use if pUnk has a ref-count of 0.
        """
        ptraddress = pUnk.p.value
        if ptraddress is not None:
            _CreateAgObjectLifetimeManager._add_ref_impl(pUnk)
            self.take_ownership(pUnk, False)

    def take_ownership(self, pUnk:"IUnknown", isApplication=False):
        """
        Add pUnk to the reference manager; does not call AddRef in STK.

        Use if pUnk has a ref-count of 1.
        """
        ptraddress = pUnk.p.value
        if ptraddress is not None:
            with GcDisabler():
                if isApplication:
                    self._applications.append(ptraddress)
                if ptraddress in self._ref_counts:
                    _CreateAgObjectLifetimeManager._release_impl(pUnk)
                    self.internal_add_ref(pUnk)
                else:
                    self._ref_counts[ptraddress] = 1

    def internal_add_ref(self, pUnk:"IUnknown"):
        """Increment the internal reference count of pUnk."""
        ptraddress = pUnk.p.value
        with GcDisabler():
            if ptraddress in self._ref_counts:
                self._ref_counts[ptraddress] = self._ref_counts[ptraddress] + 1

    def release(self, pUnk:"IUnknown"):
        """
        Decrements the internal reference count of pUnk.

        If the internal reference count reaches zero, calls Release in STK.
        """
        ptraddress = pUnk.p.value
        if ptraddress is not None:
            with GcDisabler():
                if ptraddress in self._ref_counts:
                    if self._ref_counts[ptraddress] == 1:
                        _CreateAgObjectLifetimeManager._release_impl(pUnk)
                        del(self._ref_counts[ptraddress])
                    else:
                        self._ref_counts[ptraddress] = self._ref_counts[ptraddress] - 1

    def release_all(self, releaseApplication=True):
        with GcDisabler():
            preserved_app_ref_counts = dict()
            while len(self._ref_counts) > 0:
                ref_count = self._ref_counts.popitem()
                ptraddress = ref_count[0]
                if not releaseApplication and ptraddress in self._applications:
                    preserved_app_ref_counts[ptraddress] = ref_count[1]
                    continue
                pUnk = IUnknown()
                pUnk.p = c_void_p(ptraddress)
                _CreateAgObjectLifetimeManager._release_impl(pUnk)
                pUnk.p = c_void_p(0)
            self._ref_counts = preserved_app_ref_counts

ObjectLifetimeManager = _CreateAgObjectLifetimeManager()


class _CreateCoInitializeManager(object):
    def __init__(self):
        self.init_count = 0

    def initialize(self):
        if self.init_count == 0:
            OLE32Lib._initialize()
            OLEAut32Lib._initialize()
            OLE32Lib.CoInitializeEx(None, COINIT_APARTMENTTHREADED)
        self.init_count = self.init_count + 1

    def uninitialize(self):
        self.init_count = self.init_count - 1
        if self.init_count == 0:
            OLE32Lib.CoUninitialize()

CoInitializeManager = _CreateCoInitializeManager()

def _initialize_embedded():
    """Initialize OLE libraries for STK plugin initialization."""
    OLE32Lib._initialize()
    OLEAut32Lib._initialize()

###############################################################################
#   Interfaces
###############################################################################
class IUnknown(object):
    _guid = "{00000000-0000-0000-C000-000000000046}"
    _vtable_offset = 0
    _num_methods = 3
    _QueryInterface = WINFUNCTYPE(HRESULT, LPVOID, POINTER(GUID), POINTER(LPVOID))
    _metadata = {
        "iid_data" : (0, 5044031582654955712),
    }
    if os.name == "nt":
        _QIIndex = 0
    else:
        _QIIndex = 2
    def __init__(self, pUnk=None):
        if pUnk is not None:
            self.p = pUnk.p
            self.add_ref()
        else:
            self.p = c_void_p()
    def __del__(self):
        if self:
            self.release()
    def __eq__(self, other):
        return self.p.value == other.p.value
    def __hash__(self):
        return self.p.value
    def __bool__(self):
        return self.p.value is not None and self.p.value > 0
    def _get_vtbl_entry(self, index):
        vptr = cast(self.p, POINTER(c_void_p))
        vtbl = cast(vptr.contents, POINTER(c_void_p))
        return vtbl[index]
    def _query_backwards_compatability_interface(self, iid):
        from .coclassutil import AgBackwardsCompatabilityMapping
        iid_tuple = iid.as_data_pair()
        if AgBackwardsCompatabilityMapping.check_guid_available(iid_tuple):
            old_iid = GUID.from_data_pair(AgBackwardsCompatabilityMapping.get_old_guid(iid_tuple))
            return self.query_interface(old_iid)
        return None
    def query_interface(self, arg:typing.Union[GUID, dict]) -> "IUnknown":
        if type(arg) == dict:
            iid = GUID.from_data_pair(arg["iid_data"])
        else:
            iid = arg
        pIntf = IUnknown()
        hr = IUnknown._QueryInterface(self._get_vtbl_entry(IUnknown._QIIndex))(self.p, byref(iid), byref(pIntf.p))
        if not Succeeded(hr):
            return self._query_backwards_compatability_interface(iid)
        pIntf.take_ownership()
        return pIntf
    def create_ownership(self):
        """Call AddRef on the pointer, and register the pointer to be Released when the ref count goes to zero."""
        ObjectLifetimeManager.create_ownership(self)
    def take_ownership(self, isApplication=False):
        """Register the pointer to be Released when the ref count goes to zero but does not call AddRef."""
        ObjectLifetimeManager.take_ownership(self, isApplication)
    def add_ref(self):
        """Increment the ref count if the pointer was registered.

        Pointer registration must be done by create_ownership or take_ownership.
        """
        ObjectLifetimeManager.internal_add_ref(self)
    def release(self):
        """Decrement the ref count if the pointer was registered. Calls Release if the ref count goes to zero.

        Pointer registration must be done by create_ownership or take_ownership.
        """
        ObjectLifetimeManager.release(self)

    def invoke(self, intf_metadata:dict, method_metadata:dict, *args):
        return self._invoke_impl(intf_metadata, method_metadata, *args)

    def get_property(self, intf_metadata:dict, method_metadata:dict):
        return self._invoke_impl(intf_metadata, method_metadata, OutArg())

    def set_property(self, intf_metadata:dict, method_metadata:dict, value):
        return self._invoke_impl(intf_metadata, method_metadata, value)

    def _invoke_impl(self, intf_metadata:dict, method_metadata:dict, *args):
        from .coclassutil import evaluate_hresult
        guid = GUID.from_data_pair(intf_metadata["iid_data"])
        method_offset = method_metadata["offset"]
        vtable_index = intf_metadata["vtable_reference"] + method_offset
        arg_types = method_metadata["arg_types"]
        marshaller_classes = method_metadata["marshallers"]
        method = IFuncType(self, guid, vtable_index, *arg_types)
        marshallers = []
        for arg, marshaller_class in zip(args, marshaller_classes):
            if type(arg) is OutArg:
                marshaller = marshaller_class()
                marshallers.append(marshaller)
            else:
                marshaller = marshaller_class(arg)
                marshallers.append(marshaller)
        call_args = []
        for marshaller, arg_type in zip(marshallers, arg_types):
            if type(arg_type) == type(POINTER(PVOID)): # type(POINTER(X)) == type(POINTER(Y)) for all X,Y. the choice of PVOID was arbitrary
                call_args.append(byref(marshaller.com_val))
            else:
                call_args.append(marshaller.com_val)
        evaluate_hresult(method(*call_args))
        return_vals = []
        for arg, marshaller in zip(args, marshallers):
            if type(arg) is OutArg:
                return_vals.append(marshaller.python_val)
        del(marshallers)
        if len(return_vals) == 0:
            return
        elif len(return_vals) == 1:
            return return_vals[0]
        else:
            return tuple(return_vals)

class IDispatch(IUnknown):
    _guid = "{00020400-0000-0000-C000-000000000046}"
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _num_methods = 4
    def __init__(self, pUnk: IUnknown):
        IUnknown.__init__(self, pUnk)

class IPictureDisp(IUnknown):
    def __init__(self):
        raise STKRuntimeError("IPictureDisp not supported.")

class IEnumVariant(object):
    guid = "{00020404-0000-0000-C000-000000000046}"
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _num_methods = 4
    def __init__(self, pUnk):
        self.pUnk = pUnk
        IID_IEnumVARIANT = GUID(IEnumVariant.guid)
        vtable_offset_local = IEnumVariant._vtable_offset - 1
        self._Next  = IFuncType(pUnk, IID_IEnumVARIANT, vtable_offset_local+1, ULONG, POINTER(Variant), POINTER(ULONG))
        self._Reset = IFuncType(pUnk, IID_IEnumVARIANT, vtable_offset_local+3)
    def next(self) -> Variant:
        from .marshall import python_val_from_VARIANT
        one_obj = ULONG(1)
        num_fetched = ULONG()
        obj = Variant()
        OLEAut32Lib.VariantInit(obj)
        if self._Next(one_obj, byref(obj), byref(num_fetched)) == S_OK:
            return python_val_from_VARIANT(obj, clear_variant=True)
        else:
            return None
    def reset(self) -> None:
        self._Reset()

class IFuncType(object):
    """Wrapper for calling methods into COM interface vtables."""
    def __init__(self, pUnk, iid, method_index, *argtypes):
        self.pUnk = pUnk
        self.iid = iid
        self.index = method_index
        self.method = WINFUNCTYPE(HRESULT, LPVOID, *argtypes)
    def __call__(self, *args):
        pIntf: IUnknown = self.pUnk.query_interface(self.iid)
        ret = self.method(pIntf._get_vtbl_entry(self.index))(pIntf.p, *args)
        del(pIntf)
        return ret



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Make code edits."""

import logging
from pathlib import Path

from libcst import MetadataWrapper, parse_module

from .migration_transformer import MigrationTransformer


class CodeEditor:
    """Edit the code to apply the API migrations identified through both static and dynamic analysis."""

    def __init__(self, recording, mappings):
        """Construct a new editor configured to apply the specified recording and mappings"""
        self.recording = recording
        self.mappings = mappings

    def apply_changes(self):
        """Apply the code migrations."""
        files_to_edit = self.recording.get_files_to_edit()
        for file_to_edit in files_to_edit:
            source_code = Path(file_to_edit).read_text(encoding="utf-8")
            tree = MetadataWrapper(parse_module(source_code))

            transformer = MigrationTransformer(file_to_edit, self.recording, self.mappings)

            modified_tree = tree.visit(transformer)

            migrated_filename = file_to_edit + "-migrated"
            logging.info(f"Writing {migrated_filename}")
            Path(migrated_filename).write_text(modified_tree.code, encoding="utf-8")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for LineTarget objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import LineTarget
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import line_chart


def anchor_point_solar_az_el_line_chart(stk_object :LineTarget, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the elevation and azimuth over time, describing the relative position vector of the apparent Sun to the anchor point, with respect to the local horizontal plane.

    This frame has the Z axis aligned with the inward surface normal direction (minus Z is up) and the X axis constrained toward the local north direction.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LineTarget\\Anchor Point Solar Az-El.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LineTarget
        The STK LineTarget object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Anchor Point Lighting AER').execute_elements(start_time, stop_time, step, ['Time', 'Elevation', 'Azimuth']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Longitude', 'lines': [
            {'y_name':'azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': None, 'dimension': 'Longitude'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'elevation', 'label':'Elevation', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['azimuth','elevation'], ['time'], axes, 'time', 'Time', 'Anchor Point Solar Az-El', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)




# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Command line arguments for the API migration assistant."""

import argparse
from pathlib import Path


class ArgumentParser:
    """Parser for the command line arguments supported by the API migration assistant."""

    @staticmethod
    def create_parser(record_handler, apply_handler):
        """Create the argparse parser for the command line arguments."""
        # Create the top-level parser
        parser = argparse.ArgumentParser(
            description="API Migration Assistant",
            add_help=True,
        )

        parser.add_argument(
            "--log",
            action="store",
            dest="log_level",
            metavar="<level>",
            help="sets the level for logging (one of DEBUG, INFO [default], WARNING, ERROR)",
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        )

        subparsers = parser.add_subparsers(help="action to execute", required=True)
        ArgumentParser._add_record_sub_parser(subparsers, record_handler)
        ArgumentParser._add_apply_sub_parser(subparsers, apply_handler)

        return parser

    @staticmethod
    def _add_record_sub_parser(subparsers, record_handler):
        record_parser = subparsers.add_parser(
            "record",
            help="record the execution of the specified program to identify the API calls to migrate",
        )
        record_parser.add_argument(
            "--entry-point",
            action="store",
            default="main",
            metavar="<entry point>",
            dest="entry_point",
            help="entry point to invoke (default: main)",
        )
        record_parser.add_argument(
            "--root-directory",
            metavar="<directory>",
            action="store",
            dest="root_directory",
            help="only migrate files under this directory (default: program directory)",
        )

        ArgumentParser._add_mappings_recordings_directory_arguments(record_parser)

        record_parser.add_argument(
            "-m",
            action="store_true",
            default=False,
            dest="run_as_module",
            help="invoke the specified program as a module",
        )
        record_parser.add_argument(
            "program",
            type=str,
            help="script file or module (if -m flag) to record",
        )
        record_parser.add_argument(
            "program_args",
            metavar="...",
            help="arguments passed to program in sys.argv[1:]",
            nargs=argparse.REMAINDER,
        )
        record_parser.set_defaults(handler=lambda args: record_handler(record_parser, args))

    @staticmethod
    def _add_apply_sub_parser(subparsers, apply_handler):
        apply_parser = subparsers.add_parser("apply", help="migrate the calls previously recorded to the new API")
        ArgumentParser._add_mappings_recordings_directory_arguments(apply_parser)
        apply_parser.set_defaults(handler=lambda args: apply_handler(apply_parser, args))

    @staticmethod
    def _add_mappings_recordings_directory_arguments(parser):
        default_mappings_directory = str(
            (Path(__file__).parent / "api-mappings").resolve()
        )  # points to the mappings provided with the library
        default_recordings_directory = str((Path.cwd() / "recordings").resolve())

        parser.add_argument(
            "--mappings-directory",
            action="store",
            dest="mappings_directory",
            metavar="<directory>",
            default=default_mappings_directory,
            help=f"directory containing the JSON API mappings (default: {default_mappings_directory})",
        )
        parser.add_argument(
            "--recordings-directory",
            action="store",
            dest="recordings_directory",
            metavar="<directory>",
            default=default_recordings_directory,
            help=f"directory receiving the JSON recordings (default: {default_recordings_directory})",
        )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ["STKXInitialize"]



from ..internal  import comutil          as agcom
from ..internal  import coclassutil      as agcls
from ..internal  import marshall         as agmarshall
from ..internal.comutil     import IDispatch
from ..internal.apiutil     import (InterfaceProxy, initialize_from_source_object, get_interface_property,
    set_class_attribute, SupportsDeleteCallback)
from ..utilities.exceptions import STKRuntimeError


def _raise_uninitialized_error(*args):
    raise STKRuntimeError("Valid STK object model classes are returned from STK methods and should not be created independently.")


class STKXInitialize(SupportsDeleteCallback):
    """STK X Advanced Initialization Options."""

    _num_methods = 3
    _vtable_offset = IDispatch._vtable_offset + IDispatch._num_methods
    _initialize_activation_context_method_offset = 1
    _initialize_data_method_offset = 2
    _initialize_data_ex_method_offset = 3
    _metadata = {
        "iid_data" : (5587570431076459601, 3289199399803536798),
        "vtable_reference" : IDispatch._vtable_offset + IDispatch._num_methods - 1,
    }
    _property_names = {}
    def _get_property(self, attrname):
        return get_interface_property(attrname, STKXInitialize)

    _initialize_activation_context_metadata = { "offset" : _initialize_activation_context_method_offset,
            "arg_types" : (),
            "marshallers" : () }
    def initialize_activation_context(self) -> None:
        """Initialize the activation context to be used by STK Engine based on the current activation context."""
        return self._intf.invoke(STKXInitialize._metadata, STKXInitialize._initialize_activation_context_metadata, )

    _initialize_data_metadata = { "offset" : _initialize_data_method_offset,
            "arg_types" : (agcom.BSTR, agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg, agmarshall.BStrArg,) }
    def initialize_data(self, install_home:str, config_directory:str) -> None:
        """Copy the virtual registry to the Config directory and initialize it with the install home specified."""
        return self._intf.invoke(STKXInitialize._metadata, STKXInitialize._initialize_data_metadata, install_home, config_directory)

    _initialize_data_ex_metadata = { "offset" : _initialize_data_ex_method_offset,
            "arg_types" : (agcom.BSTR, agcom.BSTR, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL, agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.BStrArg, agmarshall.BStrArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg, agmarshall.VariantBoolArg,) }
    def initialize_data_ex(self, install_home:str, config_directory:str, defaults:bool, styles:bool, vgt:bool, amm:bool, gator:bool, online_data:bool, online_sgp4:bool) -> None:
        """Copy the virtual registry to the Config directory and initialize it with the install home specified, and config options."""
        return self._intf.invoke(STKXInitialize._metadata, STKXInitialize._initialize_data_ex_metadata, install_home, config_directory, defaults, styles, vgt, amm, gator, online_data, online_sgp4)


    def __init__(self, source_object=None):
        """Construct an object of type STKXInitialize."""
        SupportsDeleteCallback.__init__(self)
        initialize_from_source_object(self, source_object, STKXInitialize)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, STKXInitialize, [STKXInitialize, ])

agcls.AgClassCatalog.add_catalog_entry((5130722036779683869, 8760598985969493655), STKXInitialize)
agcls.AgTypeNameMap["STKXInitialize"] = STKXInitialize


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
PySTK Scenario Construction Extensions.

A set of convenience utilities to facilitate the construction of scenarios.
"""

from ansys.stk.core.stkobjects import ISTKObject, Scenario, STKObjectRoot, STKObjectType


def construct_scenario_from_dict(root: STKObjectRoot, dictionary: dict) -> Scenario:
    """Create a scenario with sub-objects from a dictionary describing its structure.

    The dictionary represents a hierarchy of nodes, where each node has a type, a name, and children. The type
    is case insensitive and corresponds to the values in the STKObjectType enumeration.

    Examples
    --------
    Construct a scenario with a place and a sensor attached to it:
    >>> scenario = construct_scenario_from_dict(
    >>>     root,
    >>>     {
    >>>         "type": "Scenario",
    >>>         "name": "MyScenario",
    >>>         "children": [
    >>>             {
    >>>                 "type": "Place",
    >>>                 "name": "MyPlace",
    >>>                 "children": [
    >>>                     {
    >>>                         "type": "Sensor",
    >>>                         "name": "MySensor",
    >>>                         "children": []
    >>>                     }
    >>>                  ]
    >>>             }
    >>>         ]
    >>>     })

    Parameters
    ----------
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    dict : dict
        A dictionary describing the scenario to construct.

    Returns
    -------
    ansys.stk.core.stkobjects.Scenario
        The newly created scenario.

    Raises
    ------
    RuntimeError
        If the dictionary does not have the proper structure.
    """

    def _validate_node(node):
        if "name" not in node:
            raise RuntimeError("Unexpected dictionary structure, missing 'name'")

        if "type" not in node:
            raise RuntimeError("Unexpected dictionary structure, missing 'type'")

        if "children" not in node:
            raise RuntimeError("Unexpected dictionary structure, missing 'children'")

        if len(node) != 3:
            raise RuntimeError("Unexpected dictionary structure, expected 3 entries")

    def _create_object(parent: ISTKObject, node: dict):
        for child in node["children"]:
            _validate_node(child)
            new_object_type_name: str = child["type"].upper()
            if new_object_type_name in STKObjectType.__dict__:
                new_object_type: STKObjectType = STKObjectType.__dict__[new_object_type_name]
                new_object_name: str = child["name"]
                new_object = parent.children.new(new_object_type, new_object_name)
                _create_object(new_object, child)
            else:
                raise RuntimeError(f"{new_object_type_name} is not a valid STK object type")

    if root.current_scenario is not None:
        root.close_scenario()

    _validate_node(dictionary)

    root.new_scenario(dictionary["name"])
    scenario: ISTKObject = root.current_scenario
    _create_object(parent=scenario, node=dictionary)

    return Scenario(root.current_scenario)
